"""
    @file:              patient_data_generator.py
    @Author:            Maxence Larose

    @Creation Date:     10/2021
    @Last modification: 01/2022

    @Description:       This file contains the PatientDataGenerator class which is used to iterate on multiple
                        patients' dicom files and segmentation files using the PatientDataReader to obtain all patients'
                        data. The PatientDataGenerator inherits from the Generator abstract class.
"""

from collections.abc import Generator
from copy import deepcopy
import json
import logging
from typing import Dict, List, NamedTuple, Optional, Union

from ..data_readers.patient_data.patient_data_reader import PatientDataReader
from ..data_model import PatientDataModel


class PatientDataGenerator(Generator):
    """
    A class used to iterate on multiple patients' dicom files and segmentation files using the PatientDataReader to
    obtain all patients' data. The PatientDataGenerator inherits from the Generator abstract class.
    """

    class PathsToPatientFolderAndSegmentations(NamedTuple):
        """
        Namedtuple of paths to patient folder and segmentations.
        """
        path_to_patient_folder: str
        path_to_segmentations: Optional[List[str]] = None

    def __init__(
            self,
            paths_to_patients_folder_and_segmentations: List[PathsToPatientFolderAndSegmentations],
            verbose: bool,
            organs: Optional[Union[str, Dict[str, List[str]]]] = None,
            series_descriptions: Optional[Union[str, Dict[str, List[str]]]] = None,
    ):
        """
        Used to check if either the series descriptions or the path to the series description json dictionary is None.

        Parameters
        ----------
        paths_to_patients_folder_and_segmentations : List[PathsToPatientFolderAndSegmentations]
            List of named tuples including the path to the folder containing the patient dicom files and the path to
            the segmentations related to these dicom files.
        verbose : bool
            True to log/print some information else False.
        organs : Optional[Union[str, Dict[str, List[str]], None]], default = None.
            A dictionary that contains the organs and their associated segment names. Keys are arbitrary organ names
            and values are lists of possible segment names. It can also be specified as a path to a json file that
            contains the organs dictionary.
        series_descriptions : Optional[Union[str, Dict[str, List[str]], None]], default = None.
            A dictionary that contains the series descriptions of the images that absolutely needs to be extracted from
            the patient's file. Keys are arbitrary names given to the images we want to add and values are lists of
            series descriptions. The images associated with these series descriptions do not need to have a
            corresponding segmentation. In fact, the whole point of adding a way to specify the series descriptions that
            must be added to the dataset is to be able to add images without their segmentation. Can be specified as a
            path to a json file that contains the series descriptions dictionary.
        """
        self.paths_to_patients_folder_and_segmentations = paths_to_patients_folder_and_segmentations

        if isinstance(series_descriptions, str):
            self.path_to_series_description_json = series_descriptions
        elif isinstance(series_descriptions, dict):
            self.series_descriptions = series_descriptions
            self.path_to_series_description_json = None
        elif series_descriptions is None:
            self.series_descriptions = None
            self.path_to_series_description_json = None
        else:
            raise TypeError(f"Given series descriptions {series_descriptions} doesn't have the right type. Allowed"
                            f" types are str, dict and None.")

        if isinstance(organs, str):
            self.path_to_organs_json = organs
        elif isinstance(organs, dict):
            self.organs = organs
            self.path_to_organs_json = None
        elif organs is None:
            if any(paths.path_to_segmentations for paths in self.paths_to_patients_folder_and_segmentations):
                raise AssertionError("The variable organs is required. It is a dictionary where keys are arbitrary"
                                     " organ names and values are lists of possible segment names.")
            self.organs = None
            self.path_to_organs_json = None
        else:
            raise TypeError(f"Given organs {organs} doesn't have the right type. Allowed types are str, dict and None.")

        self._verbose = verbose
        self.current_index = 0

        if verbose:
            logging.info(f"\n# {'-'*111} #\n# {' ' * 40} DOWNLOADING ALL PATIENTS DATA {' ' * 40} #\n# {'-'*111} #")

    def __len__(self) -> int:
        """
        Total number of patients.

        Returns
        _______
        length: int
            Total number of patients.
        """
        return len(self.paths_to_patients_folder_and_segmentations)

    @property
    def series_descriptions(self) -> Dict[str, List[str]]:
        """
        Series descriptions.

        Returns
        -------
        series_descriptions : Dict[str, List[str]]
            A dictionary that contains the series descriptions of the images that absolutely needs to be extracted from
            the patient's file. Keys are arbitrary names given to the images we want to add and values are lists of
            series descriptions. The images associated with these series descriptions do not need to have a
            corresponding segmentation. In fact, the whole point of adding a way to specify the series descriptions that
            must be added to the dataset is to be able to add images without segmentation.
        """
        return self._series_descriptions

    @series_descriptions.setter
    def series_descriptions(self, series_descriptions: Dict[str, List[str]]) -> None:
        """
        Series descriptions setter.

        Parameters
        ----------
        series_descriptions : Dict[str, List[str]]
            A dictionary that contains the series descriptions of the images that absolutely needs to be extracted from
            the patient's file. Keys are arbitrary names given to the images we want to add and values are lists of
            series descriptions. The images associated with these series descriptions do not need to have a
            corresponding segmentation. In fact, the whole point of adding a way to specify the series descriptions that
            must be added to the dataset is to be able to add images without segmentation.
        """
        items = list(series_descriptions.items())
        for previous_items, current_items in zip(items, items[1:]):
            set_intersection = set(previous_items[1]) & set(current_items[1])

            if bool(set_intersection):
                raise AssertionError(f"\nThe dictionary of series descriptions should not contain the same series names"
                                     f" for different images/modalities. \nHowever, here we find the series names "
                                     f"{previous_items[1]} for the {previous_items[0]} image and {current_items[1]} "
                                     f"for the {current_items[0]} image. \nClearly, the images series values are "
                                     f"overlapping because of the series named {set_intersection}.")

        self._series_descriptions = series_descriptions

    @property
    def path_to_series_description_json(self) -> Union[str, None]:
        """
        Path to series description json.

        Returns
        -------
        path_to_series_description_json : Union[str, None]
            Path to the json dictionary that contains the series descriptions.
        """
        return self._path_to_series_description_json

    @path_to_series_description_json.setter
    def path_to_series_description_json(self, path_to_series_description_json: str) -> None:
        """
        Path to series description json setter. This is used to set the series descriptions attribute using the
        dictionary read from the json file.

        Parameters
        ----------
        path_to_series_description_json : str
            Path to the json dictionary that contains the series descriptions.
        """
        with open(path_to_series_description_json, "r") as json_file:
            self.series_descriptions = deepcopy(json.load(json_file))

        self._path_to_series_description_json = path_to_series_description_json

    def save_series_descriptions_to_json(self, path: str) -> None:
        """
        Save the dictionary of series descriptions in a json format at the given path.

        Parameters
        ----------
        path : str
            Path to the json dictionary that will contain the series descriptions.
        """
        with open(path, 'w', encoding='utf-8') as json_file:
            json.dump(self.series_descriptions, json_file, ensure_ascii=False, indent=4)

    @property
    def organs(self) -> Dict[str, List[str]]:
        """
        Organs property.

        Returns
        -------
        organs : Dict[str, List[str]]
            A dictionary that contains the organs and their associated segment names. Keys are arbitrary organ names
            and values are the list of possible segment names for each organ.
        """
        return self._organs

    @organs.setter
    def organs(self, organs: Dict[str, List[str]]) -> None:
        """
        Organs setter.

        Parameters
        ----------
        organs : Dict[str, List[str]]
            A dictionary that contains the organs and their associated segment names. Keys are arbitrary organ names
            and values are the list of possible segment names for each organ.
        """
        self._organs = organs

    @property
    def path_to_organs_json(self) -> str:
        """
        Path to organs json property.

        Returns
        -------
        path_to_organs_json : str
            Path to a json dictionary that contains the list of possible segment names for each organ. Keys are
            arbitrary organ names and values are the list of possible segment names for each organ.
        """
        return self._path_to_series_description_json

    @path_to_organs_json.setter
    def path_to_organs_json(self, path_to_organs_json: str) -> None:
        """
        Path to organs json setter.

        Parameters
        ----------
        path_to_organs_json : str
            Path to a json dictionary that contains the list of possible segment names for each organ. Keys are
            arbitrary organ names and values are the list of possible segment names for each organ.
        """
        with open(path_to_organs_json, "r") as json_file:
            self.organs = deepcopy(json.load(json_file))

        self._path_to_organs_json = path_to_organs_json

    def send(self, _) -> PatientDataModel:
        """
        Resumes the execution and sends a value into the generator function. This method returns the next value yielded
        by the generator and update the current index or raises StopIteration (via the self.throw method) if all patient
        datasets have been generated.

        Returns
        -------
        patient_dataset : PatientDataModel
            A named tuple grouping the patient's data extracted from its dicom files and the patient's medical image
            segmentation data extracted from the segmentation files.
        """
        if self.current_index == self.__len__():
            self.throw()

        path_to_dicom_folder = self.paths_to_patients_folder_and_segmentations[self.current_index][0]
        paths_to_segmentations = self.paths_to_patients_folder_and_segmentations[self.current_index][1]

        if self._verbose:
            logging.info(f"\n\n# {'-'*50} Patient {self.current_index + 1} {'-'*50} #")

        patient_data_reader = PatientDataReader(
            path_to_dicom_folder=path_to_dicom_folder,
            verbose=self._verbose,
            organs=self.organs,
            paths_to_segmentations=paths_to_segmentations,
            series_descriptions=self.series_descriptions
        )
        self.series_descriptions = patient_data_reader.series_descriptions
        if self.path_to_series_description_json:
            self.save_series_descriptions_to_json(path=self._path_to_series_description_json)
        self.current_index += 1

        return patient_data_reader.get_patient_dataset()

    def throw(self, typ=StopIteration, value=None, traceback=None) -> None:
        """
        Raises an exception of type typ.
        """
        raise typ



"""
    @file:              base_nrrd_segmentation_factory.py
    @Author:            Maxence Larose

    @Creation Date:     10/2021
    @Last modification: 01/2022

    @Description:       This file contains the abstract class NrrdSegmentationFactory and all factories that inherit
                        from this class.
"""

from abc import abstractmethod
from collections import OrderedDict
import re
from typing import Dict, List, NamedTuple, Tuple

import nrrd
import numpy as np

from .segment import Segment
from .base_segmentation_factory import BaseSegmentationFactory


class SegmentData(NamedTuple):
    """
    A named tuple grouping all the important information of a specific segment in the nrrd segmentation.

    Elements
    --------
    name : str
        The segment name.
    layer : int
        The layer of the segment in the 4D mask array.
    label_value : int
        The label value of the segment in the 3D mask array.
    """
    name: str = None
    layer: int = None
    label_value: int = None


class BaseNrrdSegmentationFactory(BaseSegmentationFactory):
    """
    An abstract class used as a reference for all other segmentation classes that read data from files with a ".nrrd"
    extension. These classes differ by the type of segmentation they are able to read. Indeed, the format of the
    segmentations may be different even if the file extension is the same, because there is multiple ways of
    representing a segment.
    """

    def __init__(
            self,
            path_to_segmentation: str,
            organs: Dict[str, List[str]]
    ):
        """
        Used to load the segmentation data from the path to segmentation.

        Parameters
        ----------
        path_to_segmentation : str
            The path to the segmentation file.
        organs : Dict[str, List[str]]
            A dictionary that contains the organs and their associated segment names. Keys are arbitrary organ names
            and values are lists of possible segment names.

        Attributes
        ----------
        self._segmentation_data : Tuple[np.ndarray, OrderedDict]
            Tuple containing the segmentation array and header.
        """
        super(BaseNrrdSegmentationFactory, self).__init__(
            path_to_segmentation=path_to_segmentation,
            organs=organs
        )

        self._segmentation_data: Tuple[np.ndarray, OrderedDict] = nrrd.read(filename=path_to_segmentation)

    @property
    @abstractmethod
    def number_of_segments(self) -> int:
        """
        Number of segments in the segmentation.

        Returns
        -------
        number_of_segments : int
            Number of segments
        """
        pass

    @abstractmethod
    def _get_segment_from_segment_idx(self, segment_idx: int) -> SegmentData:
        """
        Get the segment data by using its index in the segment list found in the segmentation file.

        Parameters
        ----------
        segment_idx : int
            Segment index. The possible values of the segment index are [0, 1, ..., self.number_of_segments].

        Returns
        -------
        segment_data : SegmentData
            The segment data.
        """
        pass

    @property
    def segmentation_array(self) -> np.ndarray:
        """
        Segmentation array.

        Returns
        -------
        segmentation_array : np.ndarray
            Segmentation as a 3D or 4D array.
        """
        return self._segmentation_data[0]

    @property
    def _segmentation_header(self) -> OrderedDict:
        """
        Segmentation header.

        Returns
        -------
        segmentation_header : OrderedDict
            SimpleITK segmentation header.
        """
        return self._segmentation_data[1]

    @property
    def _segments_data(self) -> List[SegmentData]:
        """
        List of all the segments data in the segmentation.

        Returns
        -------
        _segments : List[SegmentData]
            List of all the segments data in the segmentation.
        """
        segments = []
        for segment_idx in range(self.number_of_segments):
            segment = self._get_segment_from_segment_idx(segment_idx=segment_idx)
            segments.append(segment)

        return segments

    @property
    def _segments_names(self) -> List[str]:
        """
        List of all the segment names.

        Returns
        -------
        _segments_names : List[str]
            List of all the segment names.
        """
        segments_names = [segment_data.name for segment_data in self._segments_data]

        return segments_names

    @property
    def _segments_names_associated_to_their_respective_organ(self) -> Dict[str, str]:
        """
        A dictionary of all the segment names associated to their respective organ.

        Returns
        -------
        segments_names_associated_to_their_respective_organ : Dict[str, str]
            A dictionary of all the segment names associated to their respective organ. Keys are segment names and
            values are organ names.
        """
        segment_names_associated_to_their_respective_organ = {}
        for segment_name in self._segments_names:
            for organ, segment_names in self._organs.items():
                if segment_name in segment_names:
                    segment_names_associated_to_their_respective_organ[segment_name] = organ
            if segment_name not in segment_names_associated_to_their_respective_organ:
                raise ValueError(f"Segment name {segment_name} doesn't correspond to any organ in the {self._organs}"
                                 f" dictionary, which are {list(self._organs.keys())}")

        return segment_names_associated_to_their_respective_organ

    def _get_the_label_map_of_a_specific_segment(self, segment: SegmentData) -> np.ndarray:
        """
        Get the binary label map associated to a given segment.

        Parameters
        ----------
        segment : SegmentData
            Segment data.

        Returns
        -------
        label_map : np.ndarray
            Binary label map of a specific segment.
        """
        if self.segmentation_array.ndim == 3:
            array = self.segmentation_array
        elif self.segmentation_array.ndim == 4:
            array = self.segmentation_array[segment.layer]
        else:
            raise ValueError(f"Segmentation array must be of dimension 3 or 4. Given array is of dimension "
                             f"{self.segmentation_array.ndim}.")

        label_map = np.zeros_like(array)
        segment_values_idx = np.where(array == segment.label_value)
        label_map[segment_values_idx] = 1

        return label_map

    @property
    def segments(self) -> List[Segment]:
        """
        Segments property.

        Returns
        -------
        segments : List[Segment]
            List of all the segments.
        """
        segments = []
        for segment in self._segments_data:
            organ_name = self._segments_names_associated_to_their_respective_organ[segment.name]
            label_map = self._get_the_label_map_of_a_specific_segment(segment=segment)

            segments.append(Segment(name=organ_name, label_map=label_map))

        return segments


class NrrdLabelMapVolumeFactory(BaseNrrdSegmentationFactory):
    """
    Class that defined the methods that are used to get the total number of segments and get the segment data
    corresponding to a given segment index for the label map volume type of segmentation.
    """

    def __init__(
            self,
            path_to_segmentation: str,
            organs: Dict[str, List[str]]
    ):
        """
        Used to load the segmentation data from the path to segmentation.

        Parameters
        ----------
        path_to_segmentation : str
            The path to the segmentation file.
        organs : Dict[str, List[str]]
            A dictionary that contains the organs and their associated segment names. Keys are arbitrary organ names
            and values are lists of possible segment names.

        Attributes
        ----------
        self._segmentation_data : Tuple[np.ndarray, OrderedDict]
            Tuple containing the segmentation array and header.
        """
        super(NrrdLabelMapVolumeFactory, self).__init__(
            path_to_segmentation=path_to_segmentation,
            organs=organs
        )

    @property
    def number_of_segments(self) -> int:
        """
        Number of segments in the segmentation.

        Returns
        -------
        number_of_segments : int
            Number of segments
        """
        return int(len(np.unique(self.segmentation_array)) - 1)

    def _get_segment_from_segment_idx(self, segment_idx: int) -> SegmentData:
        """
        Get the segment data by using its index in the segment list found in the segmentation metadata.

        Parameters
        ----------
        segment_idx : int
            Segment index. The possible values of the segment index are [0, 1, ..., self.number_of_segments].

        Returns
        -------
        segment_data : SegmentDataModel
            The segment data.
        """
        segment = SegmentData(
            name=f"Segment_{segment_idx + 1}",
            layer=0,
            label_value=segment_idx + 1
        )

        return segment


class NrrdSegmentationLabelMapRepresentationFactory(BaseNrrdSegmentationFactory):
    """
    Class that defined the methods that are used to get the total number of segments and get the segment data
    corresponding to a given segment index for the segmentation label map representation type of segmentation.
    """

    def __init__(
            self,
            path_to_segmentation: str,
            organs: Dict[str, List[str]]
    ):
        """
        Used to load the segmentation data from the path to segmentation.

        Parameters
        ----------
        path_to_segmentation : str
            The path to the segmentation file.
        organs : Dict[str, List[str]]
            A dictionary that contains the organs and their associated segment names. Keys are arbitrary organ names
            and values are lists of possible segment names.

        Attributes
        ----------
        self._segmentation_data : Tuple[np.ndarray, OrderedDict]
            Tuple containing the segmentation array and header.
        """
        super(NrrdSegmentationLabelMapRepresentationFactory, self).__init__(
            path_to_segmentation=path_to_segmentation,
            organs=organs
        )

    @property
    def number_of_segments(self) -> int:
        """
        Number of segments in the segmentation.

        Returns
        -------
        _number_of_segments : int
            Number of segments
        """
        r = re.compile(r"Segment\d+_ID")
        return int(np.sum([1 if r.match(key) else 0 for key in list(self._segmentation_header.keys())]))

    def _get_segment_from_segment_idx(self, segment_idx: int) -> SegmentData:
        """
        Get the segment data by using its index in the segment list found in the segmentation metadata.

        Parameters
        ----------
        segment_idx : int
            Segment index. The possible values of the segment index are [0, 1, ..., self.number_of_segments].

        Returns
        -------
        segment_data : SegmentDataModel
            The segment data.
        """
        segment = SegmentData(
            name=self._segmentation_header[f"Segment{segment_idx}_Name"],
            layer=int(self._segmentation_header[f"Segment{segment_idx}_Layer"]),
            label_value=int(self._segmentation_header[f"Segment{segment_idx}_LabelValue"])
        )

        return segment



"""
    @file:              segment.py
    @Author:            Maxence Larose

    @Creation Date:     12/2021
    @Last modification: 12/2021

    @Description:       This file contains the Segment class.
"""

import numpy as np


class Segment:
    """
    A class used to represent a segment object.
    """

    def __init__(self, name: str, label_map: np.ndarray):
        """
        Constructor of the segment object.

        Parameters
        ----------
        name : str
            Segment's name.
        label_map : np.ndarray
            A binary label map with the same size as the original image.
        """
        self._name = name
        self._label_map = label_map

    @property
    def name(self) -> str:
        """
        Name property.

        Returns
        -------
        name : str
            Segment's name.
        """
        return self._name

    @property
    def label_map(self) -> np.ndarray:
        """
        Label map property.

        Returns
        -------
        label_map : np.ndarray
            A binary label map with the same size as the original image.
        """
        return self._label_map



import os

"""
This file essentially gives the same important folders and paths root names to all files.
"""

ROOT = os.path.abspath(os.path.dirname(__file__))


class DatasetName:
    PATIENT_DATASET_BASE_NAME = "patient_dataset"
    RADIOMICS_DATASET_BASE_NAME = "radiomics_dataset"


class FolderName:
    DATA_FOLDER_NAME = "data"
    PATIENTS_FOLDER_NAME = "Patients"
    SEGMENTATIONS_FOLDER_NAME = "Segmentations"
    IMAGES_FOLDER_NAME = "IMAGES"


class PathName:
    PATH_TO_DATA_FOLDER = os.path.join(ROOT, FolderName.DATA_FOLDER_NAME)
    PATH_TO_PATIENTS_FOLDER = os.path.join(PATH_TO_DATA_FOLDER, FolderName.PATIENTS_FOLDER_NAME)
    PATH_TO_SEGMENTATIONS_FOLDER = os.path.join(PATH_TO_DATA_FOLDER, FolderName.SEGMENTATIONS_FOLDER_NAME)

    PATH_TO_PATIENT_DATASET = os.path.join(PATH_TO_DATA_FOLDER, DatasetName.PATIENT_DATASET_BASE_NAME)
    BASE_PATH_TO_RADIOMICS_DATASET = os.path.join(
        PATH_TO_DATA_FOLDER,
        DatasetName.RADIOMICS_DATASET_BASE_NAME
    )



"""
    @file:              patient_data_query_context.py
    @Author:            Maxence Larose

    @Creation Date:     01/2022
    @Last modification: 01/2022

    @Description:       This file contains the class PatientDataQueryContext that is used as a context class where strategies are
                        types of requests the client could ask the PatientDataReader class.
"""

from typing import Dict, List

from ...data_model import ImageDataModel, PatientDataModel
from .patient_data_query_strategy import PatientDataQueryStrategy, PatientDataQueryStrategies


class PatientDataQueryContext:
    """
    A class used as a context class where strategies are types of requests the client could make to the
    PatientDataReader class to get the patient's data.
    """

    def __init__(
            self,
            images_data: List[ImageDataModel],
            organs: Dict[str, List[str]],
            paths_to_segmentations: List[str],
            series_descriptions: Dict[str, List[str]]
    ):
        """
        Constructor of the PatientDataQueryContext class.

        Parameters
        ----------
        images_data : List[ImageDataModel]
            A list of the patient's images data.
        organs : Dict[str, List[str]]
            A dictionary that contains the organs and their associated segment names. Keys are arbitrary organ names
            and values are lists of possible segment names.
        paths_to_segmentations : List[str]
            A list of paths to the segmentation files. The name of the segmentation files must include the series uid
            of their corresponding image, i.e. the image on which the segmentation was made.
        series_descriptions : Dict[str, List[str]]
            A dictionary that contains the series descriptions of the images that absolutely needs to be extracted from
            the patient's file. Keys are arbitrary names given to the images we want to add and values are lists of
            series descriptions. The images associated with these series descriptions do not need to have a
            corresponding segmentation. In fact, the whole point of adding a way to specify the series descriptions that
            must be added to the dataset is to be able to add images without segmentation.
        """
        self._images_data = images_data
        self._organs = organs
        self._paths_to_segmentations = paths_to_segmentations
        self._series_descriptions = series_descriptions

    @property
    def patient_data_query_strategy(self) -> PatientDataQueryStrategy:
        """
        Patient data query strategy corresponding to the given paths to segmentations and series descriptions
        configuration.

        Returns
        -------
        patient_data_query_strategy : PatientDataQueryStrategy
            Patient data query strategy.
        """
        if self._paths_to_segmentations and self._series_descriptions:
            return PatientDataQueryStrategies.SEGMENTATION_AND_SERIES_DESCRIPTION
        elif self._paths_to_segmentations is None and self._series_descriptions:
            return PatientDataQueryStrategies.SERIES_DESCRIPTION
        elif self._paths_to_segmentations and self._series_descriptions is None:
            return PatientDataQueryStrategies.SEGMENTATION
        else:
            return PatientDataQueryStrategies.DEFAULT

    @property
    def _patient_data_factory_instance(self) -> PatientDataQueryStrategy.factory:
        """
        The factory class instance corresponding to the class of the given patient data strategy.

        Returns
        -------
        _patient_data_factory_instance : PatientDataQueryStrategy.factory
            Factory class instance used to get a patient's data.
        """
        _patient_data_factory_instance = self.patient_data_query_strategy.factory(
            images_data=self._images_data,
            organs=self._organs,
            paths_to_segmentations=self._paths_to_segmentations,
            series_descriptions=self._series_descriptions
        )

        return _patient_data_factory_instance

    def create_patient_data(self) -> PatientDataModel:
        """
        Creates a tuple containing all the patient's data.

        Returns
        -------
        patient_data: PatientDataModel
            Patient data.
        """
        return self._patient_data_factory_instance.create_patient_data()



import logging

from .data_generators.patient_data_generator import PatientDataGenerator
from .datasets.hdf5_dataset import PatientDataset
from .logging_tools import logs_file_setup

logs_file_setup(__file__, logging.INFO)

__author__ = "Maxence Larose"
__version__ = "0.1.1"
__copyright__ = "Copyright 2022, Maxence Larose"
__credits__ = ["Maxence Larose"]
__license__ = "GPL"
__maintainer__ = "Maxence Larose"
__email__ = "maxencelarose@hotmail.com"
__status__ = "Production"



from typing import Dict, List, NamedTuple

import pydicom
import SimpleITK as sitk
import numpy as np


class SegmentationDataModel(NamedTuple):
    """
    A named tuple grouping the segmentation as several binary label maps (one for each organ in the segmentation) and
    the segmentation as a simpleITK image.

    Elements
    --------
    binary_label_maps : Dict[str, np.ndarray]
        Dictionary grouping organs to their corresponding binary label map. The keys are organ names and the values are
        the binary label maps. The binary label maps dictionary is formatted as follows :

            binary_label_maps = {
                organ_name (example: "PROSTATE"): np.ndarray,
                organ_name (example: "RECTUM"): np.ndarray,
                ...
            }
    simple_itk_label_map : Image
        The segmentation as a SimpleITK image.
    """
    binary_label_maps: Dict[str, np.ndarray] = None
    simple_itk_label_map: sitk.Image = None


class ImageDataModel(NamedTuple):
    """
    A named tuple grouping the patient's dicom header and its medical image as a simpleITK image.

    Elements
    --------
    dicom_header : FileDataset
        Dicom header dataset.
    simple_itk_image : Image
        Segmentation as a SimpleITK image.
    """
    dicom_header: pydicom.dataset.FileDataset
    simple_itk_image: sitk.Image


class ImageAndSegmentationDataModel(NamedTuple):
    """
    A named tuple grouping the patient data retrieved from his dicom files and the segmentation data retrieved from
    the segmentation file.

    Elements
    --------
    image : ImageDataModel
        The patient's medical image data.
    segmentation : SegmentationDataModel
        Data from the segmentation of the patient's medical image.
    """
    image: ImageDataModel
    segmentation: SegmentationDataModel = None


class PatientDataModel(NamedTuple):
    """
    A named tuple grouping the patient's data extracted from its dicom files and the patient's medical image
    segmentation data extracted from the segmentation files, for each available modality. The patient data model is
    formatted as follows :

        PatientDataModel = (
            "patient_name": str,
            "data": [
                ImageAndSegmentationDataModel,
                ImageAndSegmentationDataModel
                ...
            ]
        )
    """
    patient_name: str
    data: List[ImageAndSegmentationDataModel]



import enum
from typing import NamedTuple, Callable

from .factories.nrrd_segmentation_factories import NrrdLabelMapVolumeFactory, \
    NrrdSegmentationLabelMapRepresentationFactory


class SegmentationStrategy(NamedTuple):
    name: str
    file_extension: str
    factory: Callable


class SegmentationStrategies(SegmentationStrategy, enum.Enum):

    LABEL_MAP_VOLUME: SegmentationStrategy = SegmentationStrategy(
        name="label_map_volume",
        file_extension=".nrrd",
        factory=NrrdLabelMapVolumeFactory
    )

    SEGMENTATION_LABEL_MAP_REPRESENTATION: SegmentationStrategy = SegmentationStrategy(
        name="segmentation_label_map_representation",
        file_extension=".seg.nrrd",
        factory=NrrdSegmentationLabelMapRepresentationFactory
    )



"""
    @file:              segmentation.py
    @Author:            Maxence Larose

    @Creation Date:     12/2021
    @Last modification: 12/2021

    @Description:       This file contains the class Segmentation.
"""

from typing import Dict, List

import numpy as np

from .segment import Segment


class Segmentation:
    """
    A class used as a reference for all other abstract segmentation classes. These classes will differ by the extension
    of the segmentation files they are able to read.
    """

    def __init__(self, segments: List[Segment]):
        """
        Constructor of the segmentation object.

        Parameters
        ----------
        segments : List[Segment]
            List of all the segments.
        """
        self._segments = segments

    def __len__(self) -> int:
        """
        Number of segments in the segmentation.

        Returns
        -------
        number_of_segments : int
            Number of segments
        """
        return len(self._segments)

    def __getitem__(self, index: int) -> Segment:
        """
        Get a specific segment using its index.

        Parameters
        ----------
        index : int
            Segment index. The possible values of the segment index are [0, 1, ..., self.number_of_segments].

        Returns
        -------
        segment_data : SegmentDataModel
            The segment data.
        """
        return self._segments[index]

    def __setitem__(self, index: int, value) -> None:
        """
        Set a segment using its index and its value.

        Parameters
        ----------
        index : int
            Segment index. The possible values of the segment index are [0, 1, ..., self.number_of_segments].

        Returns
        -------
        segment_data : SegmentDataModel
            The segment data.
        """
        self._segments[index] = value

    @property
    def label_maps(self) -> Dict[str, np.ndarray]:
        """
        Label maps of all the organs/segments that are found in the segmentation file.

        Returns
        -------
        label_maps : Dict[str, np.ndarray]
            A dictionary that contains the name of the organs and their corresponding binary label map. Keys are organ
            names and values are binary label maps. Thus, the label maps dictionary is formatted as follows :

                label_maps = {
                    organ_name (example: "PROSTATE"): np.ndarray,
                    organ_name (example: "RECTUM"): np.ndarray,
                    ...
                }
        """
        return {segment.name: segment.label_map for segment in self._segments}



"""
    @file:              patient_dataset.py
    @Author:            Maxence Larose

    @Creation Date:     10/2021
    @Last modification: 01/2022

    @Description:       This file contains the PatientDataset class that is used to interact with an hdf5 file dataset.
                        The main purpose of this class is to create an hdf5 file dataset from multiple patients dicom
                        files and their segmentation. This class also allows the user to interact with an existing hdf5
                        file dataset through queries.
"""

import logging
import os

import h5py
import json
import SimpleITK as sitk
import time
from tqdm import tqdm
from typing import Dict, List, Optional, Union

from ..data_readers.patient_data.patient_data_reader import PatientDataReader
from ..data_generators.patient_data_generator import PatientDataGenerator
from .tools.segmentation_filename_patterns_matcher import SegmentationFilenamePatternsMatcher


class PatientDataset:
    """
    This file contains the PatientDataset class that is used to interact with an hdf5 file dataset. The main purpose of
    this class is to create an hdf5 file dataset from multiple patients dicom files and their segmentation. This class
    also allows the user to interact with an existing hdf5 file dataset through queries.
    """

    def __init__(
            self,
            path_to_dataset: str,
    ):
        """
        Used to initialize the path to the dataset.

        Parameters
        ----------
        path_to_dataset : str
            Path to dataset.
        """
        self.path_to_dataset = path_to_dataset

    @property
    def path_to_dataset(self) -> str:
        """
        Path to dataset.

        Returns
        -------
        path_to_dataset : str
            Path to dataset containing modality and organ names.
        """
        return self._path_to_dataset

    @path_to_dataset.setter
    def path_to_dataset(self, path_to_dataset: str) -> None:
        """
        Path to dataset.

        Parameters
        ----------
        path_to_dataset : str
            Path to dataset.
        """
        if path_to_dataset.endswith(".h5"):
            self._path_to_dataset = path_to_dataset
        else:
            self._path_to_dataset = f"{path_to_dataset}" + ".h5"

    def _check_authorization_of_dataset_creation(
            self,
            overwrite_dataset: bool
    ) -> None:
        """
        Check if dataset's creation is allowed.

        Parameters
        ----------
        overwrite_dataset : bool
            Overwrite existing dataset.
        """
        if os.path.exists(self.path_to_dataset):
            if not overwrite_dataset:
                raise FileExistsError("The dataset already exists. You may overwrite it using "
                                      "overwrite_dataset = True.")
            else:
                logging.info(f"Overwriting dataset (HDF5 dataset file path : {self.path_to_dataset}).")
        else:
            logging.info(f"Writing dataset (HDF5 dataset file path : {self.path_to_dataset}).")

    @staticmethod
    def get_paths_to_patient(
            paths_to_patients_dicom_folder: List[str],
            path_to_segmentations_folder: str,
            patient_number_prefix: str = "Ano"
    ) -> List[PatientDataGenerator.PathsToPatientFolderAndSegmentations]:
        """
        Paths to patients dicom folder.

        Parameters
        ----------
        paths_to_patients_dicom_folder : List[str]
            List of the paths to all the patients dicom folder.
        path_to_segmentations_folder : str
            Path to the folder containing the segmentations.
        patient_number_prefix : str
            Prefix of the patient number common to all segmentations (default is 'Ano').
        """
        paths_to_patients_folder_and_segmentations = []

        for path_to_patient_dicom_folder in paths_to_patients_dicom_folder:
            patient_data_reader = PatientDataReader(
                path_to_dicom_folder=path_to_patient_dicom_folder,
                verbose=True,
            )

            segmentation_filename_patterns_matcher = SegmentationFilenamePatternsMatcher(
                path_to_segmentations_folder=path_to_segmentations_folder,
                patient_name=patient_data_reader.patient_name,
                patient_number_prefix=patient_number_prefix
            )

            paths_to_segmentations = segmentation_filename_patterns_matcher.get_absolute_paths_to_segmentation_files()

            paths_to_patient_folder_and_segmentations = PatientDataGenerator.PathsToPatientFolderAndSegmentations(
                path_to_patient_folder=path_to_patient_dicom_folder,
                path_to_segmentations=paths_to_segmentations
            )

            paths_to_patients_folder_and_segmentations.append(paths_to_patient_folder_and_segmentations)

        return paths_to_patients_folder_and_segmentations

    def create_hdf5_dataset(
            self,
            path_to_patients_folder: str,
            path_to_segmentations_folder: str,
            images_folder_name: str,
            verbose: bool,
            series_descriptions: Optional[Union[str, Dict[str, List[str]]]] = None,
            organs: Optional[Union[str, Dict[str, List[str]]]] = None,
            patient_number_prefix: str = "Ano",
            overwrite_dataset: bool = True,
    ) -> None:
        """
        Create an hdf5 file dataset from multiple patients dicom files and their segmentation and get patient's images
        from this dataset. The goal is to create an object from which it is easier to obtain patient images and their
        segmentation than separated dicom files and segmentation files.

        Parameters
        ----------
        path_to_patients_folder : str
            Patients folder path.
        path_to_segmentations_folder : str
            Images folder name.
        series_descriptions : Union[str, Dict[str, List[str]]]
            A dictionary that contains the series descriptions of the images that absolutely needs to be extracted from
            the patient's file. Keys are arbitrary names given to the images we want to add and values are lists of
            series descriptions. The images associated with these series descriptions do not need to have a
            corresponding segmentation. In fact, the whole point of adding a way to specify the series descriptions that
            must be added to the dataset is to be able to add images without their segmentation. Can be specified as a
            path to a json dictionary that contains the series descriptions.
        organs : Union[str, Dict[str, List[str]]]
            A dictionary that contains the organs and their associated segment names. Keys are arbitrary organ names
            and values are lists of possible segment names. It can also be specified as a path to a json file that
            contains the organs dictionary.
        images_folder_name : str
            Images folder name.
        verbose : bool
            True to log/print some information else False.
        patient_number_prefix : str
            Prefix of the patient number common to all segmentations (default is 'Ano').
        overwrite_dataset : bool, default = False.
            Overwrite existing dataset.
        """
        self._check_authorization_of_dataset_creation(overwrite_dataset=overwrite_dataset)

        hf = h5py.File(self.path_to_dataset, "w")

        paths_to_patients_dicom_folder = []
        for path_to_patient_folder in os.listdir(path_to_patients_folder):
            path_to_dicom_folder = os.path.join(
                path_to_patients_folder,
                path_to_patient_folder,
                images_folder_name
            )
            paths_to_patients_dicom_folder.append(path_to_dicom_folder)

        if verbose:
            logging.info("\nAssociating patient records with their corresponding image segmentation files...")
        paths_to_patients_folder_and_segmentations = self.get_paths_to_patient(
            paths_to_patients_dicom_folder=paths_to_patients_dicom_folder,
            path_to_segmentations_folder=path_to_segmentations_folder,
        )
        if verbose:
            logging.info("Done.")

        patient_data_generator = PatientDataGenerator(
            paths_to_patients_folder_and_segmentations=paths_to_patients_folder_and_segmentations,
            verbose=verbose,
            series_descriptions=series_descriptions,
            organs=organs
        )

        if verbose:
            print("\nProgress:")
            time.sleep(0.001)
            progress_bar = tqdm(range(len(patient_data_generator)), unit='itr', postfix="patients")
            time.sleep(0.001)
        for patient_dataset in patient_data_generator:
            patient_name = patient_dataset.patient_name
            patient_group = hf.create_group(name=patient_name)

            for idx, patient_image_data in enumerate(patient_dataset.data):
                series_description = patient_image_data.image.dicom_header.SeriesDescription
                series_uid = patient_image_data.image.dicom_header.SeriesInstanceUID
                modality = patient_image_data.image.dicom_header.Modality

                image_array = sitk.GetArrayFromImage(patient_image_data.image.simple_itk_image)
                transposed_image_array = image_array.transpose(1, 2, 0)

                series_group = patient_group.create_group(name=str(idx))
                series_group.attrs.__setitem__(name="series_description", value=series_description)
                series_group.attrs.__setitem__(name="series_uid", value=str(series_uid))
                series_group.attrs.__setitem__(name="modality", value=modality)

                series_group.create_dataset(
                    name=f"image",
                    data=transposed_image_array
                )

                series_group.create_dataset(
                    name=f"dicom_header",
                    data=json.dumps(patient_image_data.image.dicom_header.to_json_dict())
                )

                if patient_image_data.segmentation is None:
                    pass
                else:
                    for organ, label_map in patient_image_data.segmentation.binary_label_maps.items():
                        series_group.create_dataset(
                            name=f"{organ}_label_map",
                            data=label_map
                        )

            if verbose:
                print("\nProgress:")
                time.sleep(0.001)
                progress_bar.update()
                time.sleep(0.001)

        if verbose:
            progress_bar.close()

        patient_data_generator.close()



import enum
from typing import NamedTuple, Callable

from .factories.patient_data_factories import DefaultPatientDataFactory, SegmentationPatientDataFactory, \
    SeriesDescriptionPatientDataFactory, SegmentationAndSeriesDescriptionPatientDataFactory


class PatientDataQueryStrategy(NamedTuple):
    name: str
    factory: Callable


class PatientDataQueryStrategies(PatientDataQueryStrategy, enum.Enum):

    DEFAULT: PatientDataQueryStrategy = PatientDataQueryStrategy(
        name="Default",
        factory=DefaultPatientDataFactory
    )

    SEGMENTATION: PatientDataQueryStrategy = PatientDataQueryStrategy(
        name="Segmentation",
        factory=SegmentationPatientDataFactory
    )

    SERIES_DESCRIPTION: PatientDataQueryStrategy = PatientDataQueryStrategy(
        name="Series description",
        factory=SeriesDescriptionPatientDataFactory
    )

    SEGMENTATION_AND_SERIES_DESCRIPTION: PatientDataQueryStrategy = PatientDataQueryStrategy(
        name="Segmentation and series description",
        factory=SegmentationAndSeriesDescriptionPatientDataFactory
    )



"""
    @file:              segmentation_context.py
    @Author:            Maxence Larose

    @Creation Date:     10/2021
    @Last modification: 01/2022

    @Description:       This file contains the class SegmentationContext that is used as a context class where
                        strategies are types of ways to load the segmentation data, or more precisely, types of
                        segmentation object factory.
"""

from typing import Dict, List

from .segmentation_strategy import SegmentationStrategy, SegmentationStrategies
from .factories.segmentation import Segmentation


class SegmentationContext:
    """
    A class used as a context class where strategies are types of ways to load the segmentation data, or more precisely,
    types of segmentation object factory. Strategies are entirely defined by the extension of the given file and so, by
    the path of the segmentation.
    """

    def __init__(
            self,
            path_to_segmentation: str,
            organs: Dict[str, List[str]]
    ):
        """
        Constructor of the SegmentationContext class.

        Parameters
        ----------
        path_to_segmentation : str
            The path to the segmentation file.
        organs : Dict[str, List[str]]
            A dictionary that contains the organs and their associated segment names. Keys are arbitrary organ names
            and values are lists of possible segment names.
        """
        self._path_to_segmentation = path_to_segmentation
        self._organs = organs

    @property
    def path_to_segmentation(self) -> str:
        """
        Path to segmentation property.

        Returns
        -------
        path_to_segmentation : str
            The path to the segmentation file.
        """
        return self._path_to_segmentation

    @path_to_segmentation.setter
    def path_to_segmentation(self, path_to_segmentation: str) -> None:
        """
        Path to segmentation setter.

        Parameters
        ----------
        path_to_segmentation : str
            The path to the segmentation file.
        """
        self._path_to_segmentation = path_to_segmentation

    @property
    def segmentation_strategy(self) -> SegmentationStrategy:
        """
        Segmentation strategy corresponding to the given segmentation file extension.

        Returns
        -------
        segmentation_strategy : SegmentationStrategy
            Segmentation strategy.
        """
        possible_segmentation_strategies: List[SegmentationStrategy] = []
        for segmentation_category in list(SegmentationStrategies):
            if self.path_to_segmentation.endswith(segmentation_category.file_extension):
                possible_segmentation_strategies.append(segmentation_category)

        return max(possible_segmentation_strategies, key=len)

    @property
    def _segmentation_factory_instance(self) -> SegmentationStrategy.factory:
        """
        The segmentation factory instance corresponding to the class of the given segmentation category.

        Returns
        -------
        _segmentation_factory_instance : SegmentationStrategy.factory
            Factory class instance used to get the label maps and the segmentation metadata from a segmentation file.
        """
        return self.segmentation_strategy.factory(
            path_to_segmentation=self.path_to_segmentation,
            organs=self._organs
        )

    def create_segmentation(self) -> Segmentation:
        """
        Creates a Segmentation object.

        Returns
        -------
        segmentation : Segmentation
            Segmentation.
        """
        return self._segmentation_factory_instance.create_segmentation()



import logging
from .logging_tools import logs_file_setup
import warnings

from .root import *

from .datasets.hdf5_dataset import PatientDataset


if __name__ == "__main__":
    # ----------------------------------------------------------------------------------------------------------- #
    #                                          Logs Setup                                                         #
    # ----------------------------------------------------------------------------------------------------------- #
    logs_file_setup(__file__, logging.INFO)
    warnings.filterwarnings("default", category=DeprecationWarning)

    # ----------------------------------------------------------------------------------------------------------- #
    #                                            Patient Dataset                                                  #
    # ----------------------------------------------------------------------------------------------------------- #
    dataset = PatientDataset(
        path_to_dataset=PathName.PATH_TO_PATIENT_DATASET,
    )

    dataset.create_hdf5_dataset(
        path_to_patients_folder=PathName.PATH_TO_PATIENTS_FOLDER,
        path_to_segmentations_folder=PathName.PATH_TO_SEGMENTATIONS_FOLDER,
        series_descriptions=os.path.join(PathName.PATH_TO_DATA_FOLDER, "series_descriptions.json"),
        organs=os.path.join(PathName.PATH_TO_DATA_FOLDER, "organs.json"),
        images_folder_name=FolderName.IMAGES_FOLDER_NAME,
        verbose=True,
        overwrite_dataset=True
    )



"""
    @file:              base_patient_data_factory.py
    @Author:            Maxence Larose

    @Creation Date:     01/2022
    @Last modification: 01/2022

    @Description:       This file contains the class BasePatientDataFactory that is used as an abstract class used as a
                        reference for all other patient data factories.
"""

from abc import ABC, abstractmethod
from typing import Dict, List, Optional

from ....data_model import ImageDataModel, PatientDataModel


class BasePatientDataFactory(ABC):
    """
    An abstract class that is used as a reference for all other patient data factories.
    """

    def __init__(
            self,
            images_data: List[ImageDataModel],
            organs: Dict[str, List[str]],
            paths_to_segmentations: Optional[List[str]] = None,
            series_descriptions: Optional[Dict[str, List[str]]] = None
    ):
        """
        Constructor of the class BasePatientDataFactory.

        Parameters
        ----------
        images_data : List[ImageDataModel]
            A list of the patient's images data.
        organs : Dict[str, List[str]]
            A dictionary that contains the organs and their associated segment names. Keys are arbitrary organ names
            and values are lists of possible segment names.
        paths_to_segmentations : Optional[List[str]]
            A list of paths to the segmentation files. The name of the segmentation files must include the series uid
            of their corresponding image, i.e. the image on which the segmentation was made.
        series_descriptions : Optional[Dict[str, List[str]]]
            A dictionary that contains the series descriptions of the images that absolutely needs to be extracted from
            the patient's file. Keys are arbitrary names given to the images we want to add and values are lists of
            series descriptions. The images associated with these series descriptions do not need to have a
            corresponding segmentation. In fact, the whole point of adding a way to specify the series descriptions that
            must be added to the dataset is to be able to add images without segmentation.
        """
        self._images_data = images_data
        self._organs = organs
        self._paths_to_segmentations = paths_to_segmentations
        self._series_descriptions = series_descriptions

    @property
    def patient_name(self) -> str:
        """
        Patient name.

        Returns
        -------
        patient_name : str
            Patient name.
        """
        patient_name = self._images_data[0].dicom_header.PatientName

        return str(patient_name)

    @abstractmethod
    def create_patient_data(self) -> PatientDataModel:
        """
        Creates a tuple containing all the patient's data.

        Returns
        -------
        patient_data: PatientDataModel
            Patient data.
        """
        pass



import logging


def logs_file_setup(file: str, level=logging.INFO):
    import os
    import sys
    import time
    from datetime import date

    today = date.today()
    timestamp = str(time.time()).replace('.', '')
    logs_dir = f"logs/logs-{today.strftime('%d-%m-%Y')}"
    logs_file = f'{logs_dir}/{os.path.splitext(os.path.basename(file))[0]}-{timestamp}.log'
    os.makedirs(logs_dir, exist_ok=True)
    logging.basicConfig(filename=logs_file, filemode='w+', level=level)
    handler = logging.StreamHandler(sys.stdout)
    logging.getLogger().addHandler(handler)


if __name__ == '__main__':
    logs_file_setup(__file__, logging.DEBUG)



"""
    @file:              patient_data_factories.py
    @Author:            Maxence Larose

    @Creation Date:     01/2022
    @Last modification: 01/2022

    @Description:       This file contains all factories that inherit from the BasePatientDataFactory class.
"""

from typing import Dict, List, Optional

from ....data_model import ImageDataModel, ImageAndSegmentationDataModel, PatientDataModel
from .base_patient_data_factory import BasePatientDataFactory
from ...segmentation.segmentation_reader import SegmentationReader


class DefaultPatientDataFactory(BasePatientDataFactory):
    """
    Class that defined the methods that are used to get the patient data. The default factory consists in obtaining all
    the images without any segmentation.
    """

    def __init__(
            self,
            images_data: List[ImageDataModel],
            organs: Dict[str, List[str]],
            paths_to_segmentations: Optional[List[str]] = None,
            series_descriptions: Optional[Dict[str, List[str]]] = None
    ):
        """
        Constructor of the class BasePatientDataFactory.

        Parameters
        ----------
        images_data : List[ImageDataModel]
            A list of the patient's images data.
        organs : Dict[str, List[str]]
            A dictionary that contains the organs and their associated segment names. Keys are arbitrary organ names
            and values are lists of possible segment names.
        paths_to_segmentations : Optional[List[str]]
            A list of paths to the segmentation files. The name of the segmentation files must include the series uid
            of their corresponding image, i.e. the image on which the segmentation was made.
        series_descriptions : Optional[Dict[str, List[str]]]
            A dictionary that contains the series descriptions of the images that absolutely needs to be extracted from
            the patient's file. Keys are arbitrary names given to the images we want to add and values are lists of
            series descriptions. The images associated with these series descriptions do not need to have a
            corresponding segmentation. In fact, the whole point of adding a way to specify the series descriptions that
            must be added to the dataset is to be able to add images without segmentation.
        """
        super(DefaultPatientDataFactory, self).__init__(
            images_data=images_data,
            organs=organs,
            paths_to_segmentations=paths_to_segmentations,
            series_descriptions=series_descriptions
        )

    def create_patient_data(self) -> PatientDataModel:
        """
        Creates a tuple containing all the patient's data.

        Returns
        -------
        patient_data: PatientDataModel
            Patient data.
        """
        patient_data = PatientDataModel(
            patient_name=self.patient_name,
            data=[ImageAndSegmentationDataModel(image=image) for image in self._images_data]
        )
        return patient_data


class SegmentationPatientDataFactory(BasePatientDataFactory):
    """
    Class that defined the methods that are used to get the patient data. The segmentation patient data factory consists
    in obtaining the images that have the same serial uids as those contained in the file names of the given
    segmentations. The final dataset therefore contains both the segmentations and their corresponding images.
    """

    def __init__(
            self,
            images_data: List[ImageDataModel],
            organs: Dict[str, List[str]],
            paths_to_segmentations: Optional[List[str]] = None,
            series_descriptions: Optional[Dict[str, List[str]]] = None
    ):
        """
        Constructor of the class SegmentationOnlyPatientDataFactory.

        Parameters
        ----------
        images_data : List[ImageDataModel]
            A list of the patient's images data.
        organs : Dict[str, List[str]]
            A dictionary that contains the organs and their associated segment names. Keys are arbitrary organ names
            and values are the list of possible segment names for each organ.
        paths_to_segmentations : Optional[List[str]]
            A list of paths to the segmentation files. The name of the segmentation files must include the series uid
            of their corresponding image, i.e. the image on which the segmentation was made.
        series_descriptions : Optional[Dict[str, List[str]]]
            A dictionary that contains the series descriptions of the images that absolutely needs to be extracted from
            the patient's file. Keys are arbitrary names given to the images we want to add and values are lists of
            series descriptions. The images associated with these series descriptions do not need to have a
            corresponding segmentation. In fact, the whole point of adding a way to specify the series descriptions that
            must be added to the dataset is to be able to add images without segmentation.
        """
        super(SegmentationPatientDataFactory, self).__init__(
            images_data=images_data,
            organs=organs,
            paths_to_segmentations=paths_to_segmentations,
            series_descriptions=series_descriptions
        )

    def create_patient_data(self) -> PatientDataModel:
        """
        Creates a tuple containing all the patient's data.

        Returns
        -------
        patient_data: PatientDataModel
            Patient data.
        """
        data = []
        for image in self._images_data:
            for path_to_segmentation in self._paths_to_segmentations:
                if image.dicom_header.SeriesInstanceUID in path_to_segmentation:
                    segmentation_reader = SegmentationReader(path_to_segmentation=path_to_segmentation)

                    image_and_segmentation_data = ImageAndSegmentationDataModel(
                        image=image,
                        segmentation=segmentation_reader.get_segmentation_data()
                    )

                    data.append(image_and_segmentation_data)

        patient_data = PatientDataModel(
            patient_name=self.patient_name,
            data=data
        )
        return patient_data


class SeriesDescriptionPatientDataFactory(BasePatientDataFactory):
    """
    Class that defined the methods that are used to get the patient data. The series description patient data factory
    consists in obtaining only the images that have the given series descriptions. The final dataset therefore contains
    both the segmentations and their corresponding images.
    """

    def __init__(
            self,
            images_data: List[ImageDataModel],
            organs: Dict[str, List[str]],
            paths_to_segmentations: Optional[List[str]] = None,
            series_descriptions: Optional[Dict[str, List[str]]] = None
    ):
        """
        Constructor of the class SegmentationOnlyPatientDataFactory.

        Parameters
        ----------
        images_data : List[ImageDataModel]
            A list of the patient's images data.
        organs : Dict[str, List[str]]
            A dictionary that contains the organs and their associated segment names. Keys are arbitrary organ names
            and values are the list of possible segment names for each organ.
        paths_to_segmentations : Optional[List[str]]
            A list of paths to the segmentation files. The name of the segmentation files must include the series uid
            of their corresponding image, i.e. the image on which the segmentation was made.
        series_descriptions : Optional[Dict[str, List[str]]]
            A dictionary that contains the series descriptions of the images that absolutely needs to be extracted from
            the patient's file. Keys are arbitrary names given to the images we want to add and values are lists of
            series descriptions. The images associated with these series descriptions do not need to have a
            corresponding segmentation. In fact, the whole point of adding a way to specify the series descriptions that
            must be added to the dataset is to be able to add images without segmentation.
        """
        super(SeriesDescriptionPatientDataFactory, self).__init__(
            images_data=images_data,
            organs=organs,
            paths_to_segmentations=paths_to_segmentations,
            series_descriptions=series_descriptions
        )
        
    @property
    def flatten_series_descriptions(self) -> List[str]:
        """
        Flatten series descriptions.

        Returns
        -------
        flatten_series_description : List[str]
            Series descriptions as a list instead of a dictionary.
        """
        return [val for lst in self._series_descriptions.values() for val in lst]

    def create_patient_data(self) -> PatientDataModel:
        """
        Creates a tuple containing all the patient's data.

        Returns
        -------
        patient_data: PatientDataModel
            Patient data.
        """
        data = []
        for image_idx, image in enumerate(self._images_data):
            if image.dicom_header.SeriesDescription in self.flatten_series_descriptions:
                image_data = ImageAndSegmentationDataModel(image=image)
                data.append(image_data)

        patient_data = PatientDataModel(
            patient_name=self.patient_name,
            data=data
        )
        return patient_data


class SegmentationAndSeriesDescriptionPatientDataFactory(BasePatientDataFactory):
    """
    Class that defined the methods that are used to get the patient data. The segmentation and series description
    factory consists in obtaining the images that have the same serial uids as those contained in the file names of the
    given segmentations and the images that have the given series descriptions.
    """

    def __init__(
            self,
            images_data: List[ImageDataModel],
            organs: Dict[str, List[str]],
            paths_to_segmentations: Optional[List[str]] = None,
            series_descriptions: Optional[Dict[str, List[str]]] = None
    ):
        """
        Constructor of the class SegmentationOnlyPatientDataFactory.

        Parameters
        ----------
        images_data : List[ImageDataModel]
            A list of the patient's images data.
        organs : Dict[str, List[str]]
            A dictionary that contains the organs and their associated segment names. Keys are arbitrary organ names
            and values are the list of possible segment names for each organ.
        paths_to_segmentations : Optional[List[str]]
            A list of paths to the segmentation files. The name of the segmentation files must include the series uid
            of their corresponding image, i.e. the image on which the segmentation was made.
        series_descriptions : Optional[Dict[str, List[str]]]
            A dictionary that contains the series descriptions of the images that absolutely needs to be extracted from
            the patient's file. Keys are arbitrary names given to the images we want to add and values are lists of
            series descriptions. The images associated with these series descriptions do not need to have a
            corresponding segmentation. In fact, the whole point of adding a way to specify the series descriptions that
            must be added to the dataset is to be able to add images without segmentation.
        """
        super(SegmentationAndSeriesDescriptionPatientDataFactory, self).__init__(
            images_data=images_data,
            organs=organs,
            paths_to_segmentations=paths_to_segmentations,
            series_descriptions=series_descriptions
        )

    @property
    def flatten_series_descriptions(self) -> List[str]:
        """
        Flatten series descriptions.

        Returns
        -------
        flatten_series_description : List[str]
            Series descriptions as a list instead of a dictionary.
        """
        return [val for lst in self._series_descriptions.values() for val in lst]
        
    def create_patient_data(self) -> PatientDataModel:
        """
        Creates a tuple containing all the patient's data.

        Returns
        -------
        patient_data: PatientDataModel
            Patient data.
        """
        data = []
        for image_idx, image in enumerate(self._images_data):
            image_added = False
            for path_to_segmentation in self._paths_to_segmentations:
                if image.dicom_header.SeriesInstanceUID in path_to_segmentation:
                    segmentation_reader = SegmentationReader(
                        path_to_segmentation=path_to_segmentation,
                        organs=self._organs
                    )

                    image_and_segmentation_data = ImageAndSegmentationDataModel(
                        image=image,
                        segmentation=segmentation_reader.get_segmentation_data()
                    )
                    data.append(image_and_segmentation_data)
                    image_added = True

            series_description = image.dicom_header.SeriesDescription
            if series_description in self.flatten_series_descriptions and image_added is False:
                image_data = ImageAndSegmentationDataModel(image=image)
                data.append(image_data)

        patient_data = PatientDataModel(
            patient_name=self.patient_name,
            data=data
        )
        return patient_data



"""
    @file:              segmentation_filename_patterns_matcher.py
    @Author:            Maxence Larose

    @Creation Date:     10/2021
    @Last modification: 10/2021

    @Description:       This file contains the SegmentationFilenamePatternsMatcher class whose main purpose is to obtain
                        a list of absolute paths to the segmentation files given the location of the folder containing
                        all the segmentations, the patient name and the patient number prefix used in the name of
                        segmentations file.
"""

import os
import re
from typing import List

import numpy as np


class SegmentationFilenamePatternsMatcher:
    """
    A class whose main purpose is to obtain a list of absolute paths to the segmentation files given the location of the
    folder containing all the segmentations, the patient name and the patient number prefix used in the name of
    segmentations file.
    """

    def __init__(
            self,
            path_to_segmentations_folder: str,
            patient_name: str,
            patient_number_prefix: str
    ):
        """
        Used to initialize all the class' attributes.

        Parameters
        ----------
        path_to_segmentations_folder : str
            Path to the folder containing all segmentations.
        patient_name : str
            Patient name. Must contains the patient number.
        patient_number_prefix : str
            The patient number prefix used in the filename of all segmentations.
        """
        self.path_to_segmentations_folder = path_to_segmentations_folder
        self.patient_name = patient_name
        self.patient_number_prefix = patient_number_prefix

    @property
    def patient_name(self) -> str:
        """
        Patient name.

        Returns
        -------
        patient_name : str
            Patient name. Must contains the patient number.
        """
        return self._patient_name

    @patient_name.setter
    def patient_name(
            self,
            patient_name: str
    ) -> None:
        """
        Set patient name.

        Parameters
        ----------
        patient_name : str
            Patient name. Must contains the patient number.
        """
        assert any(char.isdigit() for char in patient_name), f"Patient names must contain a number. Given name is " \
                                                             f"{patient_name}."

        self._patient_name = patient_name

    @property
    def patient_number(self) -> int:
        """
        Get patient number from patient name.

        Returns
        -------
        patient_number : int
            Patient number.
        """
        patient_number = int(re.findall(pattern=r"\d+", string=self.patient_name)[-1])

        return patient_number

    @property
    def patterns(self) -> List[str]:
        """
        Pattern in the filename.

        Returns
        -------
        pattern : str
            Pattern.
        """
        return [f"{self.patient_number_prefix}{str(self.patient_number)}"]

    @property
    def paths_to_segmentation_files(self) -> List[str]:
        """
        Paths to segmentation files.

        Returns
        -------
        paths : List[str]
            A list of all the paths to the segmentation files.
        """
        return os.listdir(self.path_to_segmentations_folder)

    @property
    def matches(self) -> List[bool]:
        """
        Get a boolean list indicating whether or not the segmentation filenames match the pattern defined using the
        patient number and the patient number prefix.

        Returns
        -------
        matches : List[bool]
            A list of booleans where the value is True if all the patterns are found in the name of the segmentation
            file and False otherwise. The indexes of the list represents a specific segmentation file (see
            paths_to_segmentation_files).
        """
        matches = [False] * len(self.paths_to_segmentation_files)

        for path_idx, path_to_segmentation_file in enumerate(self.paths_to_segmentation_files):
            if all(pattern in path_to_segmentation_file for pattern in self.patterns):
                matches[path_idx] = True

        return matches

    def get_absolute_paths_to_segmentation_files(
            self,
    ) -> List[str]:
        """
        Get the absolute paths of the segmentation files whose filenames match the pattern of the given patient name.

        Returns
        -------
        absolute_paths_to_segmentation : List[str]
            A list of the absolute paths to all the segmentation files whose filenames match the pattern of the given
            patient name.
        """
        paths_to_segmentation_file = [self.paths_to_segmentation_files[i] for i in np.where(self.matches)[0]]

        absolute_paths_to_segmentation = []
        for path_to_segmentations_file in paths_to_segmentation_file:
            absolute_path_to_segmentation = os.path.join(self.path_to_segmentations_folder, path_to_segmentations_file)
            absolute_paths_to_segmentation.append(absolute_path_to_segmentation)

        return absolute_paths_to_segmentation



"""
    @file:              segmentation_reader.py
    @Author:            Maxence Larose

    @Creation Date:     10/2021
    @Last modification: 01/2022

    @Description:       This file contains the SegmentationReader class which is used to read a given segmentation file
                        and transform its contents into the format of the SegmentationDataModel class.
"""

from typing import Dict, List

import SimpleITK as sitk

from .segmentation_context import SegmentationContext
from .factories.segmentation import Segmentation
from ...data_model import SegmentationDataModel


class SegmentationReader:
    """
    A class used to read a given segmentation file and transform its contents into the standard format of the
    SegmentationDataModel class.
    """

    def __init__(
            self,
            path_to_segmentation: str,
            organs: Dict[str, List[str]]
    ):
        """
        Constructor of the class SegmentationReader.

        Parameters
        ----------
        path_to_segmentation : str
            The path to the segmentation file.
        organs : Dict[str, List[str]]
            A dictionary that contains the organs and their associated segment names. Keys are arbitrary organ names
            and values are lists of possible segment names.
        """
        self._path_to_segmentation = path_to_segmentation
        self._organs = organs

    @property
    def __simple_itk_label_image(self) -> sitk.Image:
        """
        Simple ITK label map image.

        Returns
        -------
        simple_itk_label_map : sitk.Image
            The segmentation as a SimpleITK image.
        """
        file_reader = sitk.ImageFileReader()
        file_reader.SetFileName(fn=self._path_to_segmentation)
        simple_itk_label_map = file_reader.Execute()

        return simple_itk_label_map

    @property
    def __segmentation(self) -> Segmentation:
        """
        Creates a Segmentation object.

        Returns
        -------
        segmentation : Segmentation
            Segmentation.
        """
        segmentation_context_manager = SegmentationContext(
            path_to_segmentation=self._path_to_segmentation,
            organs=self._organs
        )

        return segmentation_context_manager.create_segmentation()

    def get_segmentation_data(self) -> SegmentationDataModel:
        """
        Get the segmentation data from the path of the segmentation file.

        Returns
        -------
        segmentation_data : SegmentationDataModel
            A named tuple grouping the segmentation as several binary label maps (one for each organ in the
            segmentation), the segmentation as a simpleITK image, and finally, metadata about the organs/segments that
            are found in the segmentation.
        """
        segmentation_data = SegmentationDataModel(
            binary_label_maps=self.__segmentation.label_maps,
            simple_itk_label_map=self.__simple_itk_label_image
        )

        return segmentation_data



import os
import enum
from typing import Callable, Union
import warnings


def check_validity_of_given_path(
        path: str
) -> None:
    """
    Raise a ValueError if the given path doesn't exist.

    Parameters
    ----------
    path : str
        A path.

    Returns
    -------
    None
    """
    if not os.path.exists(path):
        raise FileNotFoundError(f"Given path {path} does not exist.")


def check_validity_of_given_name(name: str, enum_class: Union[enum.Enum, enum.IntEnum]) -> None:
    """
    Raise a ValueError if the given anatomical plane is not a member of the AnatomicalPlane Enum Class.

    Parameters
    ----------
    name : str
        Name.
    enum_class : Union[enum.Enum, enum.IntEnum]
        Class inheriting from enum.Enum or enum.IntEnum.

    Returns
    -------
    None
    """
    if name in enum_class.__members__:
        pass
    else:
        raise ValueError(f"Given member {name} doesn't exist. Allowed members of {enum_class} are "
                         f"{list(enum_class.__members__)}.")


class Decorators:

    @classmethod
    def deprecated(cls, explanation: str) -> Callable:
        """
        Decorator to use on deprecated functions.

        Parameters
        ----------
        explanation : str
            Explanation of depreciation and recommended alternative function.

        Returns
        -------
        deprecated_func: Callable
            Deprecated function.
        """
        def deprecated_func(func):
            def wrapper(*args, **kwargs):
                warnings.warn(explanation, DeprecationWarning)
                return func(*args, **kwargs)

            return wrapper

        return deprecated_func




import setuptools

setuptools.setup(
    name="dicom2hdf",
    version="0.0.4",
    author="Maxence Larose",
    author_email="maxence.larose.1@ulaval.ca",
    description="Medical data formatting and pre-processing module whose main objective"\
                " is to build an HDF5 dataset containing all medical images of patients"\
                " (DICOM format) and their associated segmentations. The HDF5 dataset"\
                " is then easier to use to perform tasks on the medical data, such as"\
                " machine learning tasks.",
    long_description=open('README.md').read(),
    long_description_content_type="text/markdown",
    url="https://github.com/MaxenceLarose/DicomToHDF5",
    license="Apache License 2.0",
    keywords='dicom hdf5 medical image segmentation pre-processing python3',
    packages=setuptools.find_packages(),
    python_requires=">=3.6",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ],
    install_requires=[
        "h5py",
        "numpy",
        "pydicom",
        "pynrrd",
        "SimpleITK",
        "tqdm"
    ],
)



"""
    @file:              patient_data_reader.py
    @Author:            Maxence Larose

    @Creation Date:     10/2021
    @Last modification: 01/2022

    @Description:       This file contains the PatientDataReader class which is used to read dicom files AND
                        segmentation files.
"""

import logging
from typing import Dict, List, Optional

from ...data_model import PatientDataModel
from ..dicom.dicom_reader import DicomReader
from .patient_data_query_context import PatientDataQueryContext


class PatientDataReader(DicomReader):
    """
    A class used to read dicom files AND segmentation files.
    """

    def __init__(
            self,
            path_to_dicom_folder: str,
            verbose: bool,
            organs: Optional[Dict[str, List[str]]] = None,
            paths_to_segmentations: Optional[List[str]] = None,
            series_descriptions: Optional[Dict[str, List[str]]] = None,
    ):
        """
        Used to check availability of given series' uid and series descriptions in the patient's dicom files.

        Parameters
        ----------
        path_to_dicom_folder : str
            Path to the folder containing the patient dicom files.
        verbose : bool
            True to log/print some information else False.
        organs : Optional[Dict[str, List[str]], None], default = None.
            A dictionary that contains the organs and their associated segment names. Keys are arbitrary organ names
            and values are lists of possible segment names.
        paths_to_segmentations : Optional[List[str]], default = None.
            A list of paths to the segmentation files. The name of the segmentation files must include the series uid
            of their corresponding image, i.e. the image on which the segmentation was made.
        series_descriptions : Optional[Dict[str, List[str]]], default = None.
            A dictionary that contains the series descriptions of the images that absolutely needs to be extracted from
            the patient's file. Keys are arbitrary names given to the images we want to add and values are lists of
            series descriptions. The images associated with these series descriptions do not need to have a
            corresponding segmentation. In fact, the whole point of adding a way to specify the series descriptions that
            must be added to the dataset is to be able to add images without segmentation.
        """
        super(PatientDataReader, self).__init__(path_to_dicom_folder=path_to_dicom_folder)
        if paths_to_segmentations:
            if any(path for path in paths_to_segmentations) and organs is None:
                raise AssertionError("The variable organs is required. It is a dictionary where keys are arbitrary"
                                     " organ names and values are lists of possible segment names.")

        self._dicom_headers = self.get_dicom_headers(verbose=verbose)
        self._organs = organs
        self._paths_to_segmentations = paths_to_segmentations
        self._series_descriptions = series_descriptions
        self._verbose = verbose

        if paths_to_segmentations is not None:
            if verbose:
                logging.info("\nChecking availability of given series uids...")
            self.check_availability_of_given_series_uids()
            if verbose:
                logging.info("Done.\n")
        if series_descriptions is not None:
            if verbose:
                logging.info("Checking availability of given series description...")
            self.check_availability_of_given_series_description()
            if verbose:
                logging.info("Done.")

    @property
    def patient_name(self) -> str:
        """
        Patient name.

        Returns
        -------
        patient_name : str
            Patient name.
        """
        return str(self._dicom_headers[0].PatientName)

    @property
    def paths_to_segmentations(self) -> List[str]:
        """
        Paths to segmentations property.

        Returns
        -------
        paths_to_segmentations : List[str]
            A list of paths to the segmentation files. The name of the segmentation files must include the series uid
            of their corresponding image, i.e. the image on which the segmentation was made.
        """
        return self._paths_to_segmentations

    @property
    def series_descriptions(self) -> Dict[str, List[str]]:
        """
        Series descriptions setter.

        Returns
        -------
        series_descriptions : Dict[str, List[str]]
            A dictionary that contains the series descriptions of the images that absolutely needs to be extracted from
            the patient's file. Keys are arbitrary names given to the images we want to add and values are lists of
            series descriptions.
        """
        return self._series_descriptions

    @series_descriptions.setter
    def series_descriptions(self, series_descriptions: Dict[str, List[str]]):
        """
        Series descriptions setter.

        Parameters
        ----------
        series_descriptions : Dict[str, List[str]]
            A dictionary that contains the series descriptions of the images that absolutely needs to be extracted from
            the patient's file. Keys are arbitrary names given to the images we want to add and values are lists of
            series descriptions. The images associated with these series descriptions do not need to have a
            corresponding segmentation. In fact, the whole point of adding a way to specify the series descriptions that
            must be added to the dataset is to be able to add images without segmentation. the patient dataset.
        """
        items = list(series_descriptions.items())
        for previous_items, current_items in zip(items, items[1:]):
            set_intersection = set(previous_items[1]) & set(current_items[1])

            if bool(set_intersection):
                raise AssertionError(f"\nThe dictionary of series descriptions should not contain the same series names"
                                     f" for different images/modalities. \nHowever, here we find the series names "
                                     f"{previous_items[1]} for the {previous_items[0]} image and {current_items[1]} "
                                     f"for the {current_items[0]} image. \nClearly, the images series values are "
                                     f"overlapping because of the series named {set_intersection}.")

        self.series_descriptions = series_descriptions

    @property
    def available_series_descriptions(self) -> List[str]:
        """
        Available series descriptions.

        Returns
        -------
        available_series_descriptions : List[str]
            Available series descriptions in the patient dicom files.
        """
        available_series_descriptions = [
            dicom_header.SeriesDescription for dicom_header in self._dicom_headers
        ]

        return available_series_descriptions + ["None"]

    @property
    def available_series_uids(self) -> List[str]:
        """
        Available series' uids.

        Returns
        -------
        available_series_uids : List[str]
            Available series uids in the patient dicom files.
        """
        available_series_uid = [
            dicom_header.SeriesInstanceUID for dicom_header in self._dicom_headers
        ]

        return available_series_uid

    def update_series_descriptions(self, series_key: str) -> None:
        """
        Add a series description to the series description list of the given series key.

        Parameters
        ----------
        series_key : str
            Series key.
        """
        logging.info(f"\nNo available series for {series_key}. \nAvailable series are "
                     f"{self.available_series_descriptions}. \nPlease write in the following location the name of the "
                     f"series to add. If the image does not require any description, write None.")

        while True:
            new_series_description = input(f"Name of the series description to add (modality = {series_key}): ")

            if new_series_description in self.available_series_descriptions:
                logging.info("Series name successfully added to the series descriptions json file.\n")
                break
            else:
                logging.info(f"The given series description name is {new_series_description}. \nHowever, this is NOT "
                             f"one of the available series description found in the patient's dicom files. \nAvailable "
                             f"series are {self.available_series_descriptions}. \nPlease try again.")

        self.series_descriptions[series_key] += [new_series_description]

    def check_availability_of_given_series_description(self) -> None:
        """
        Check availability of given series description in the patient's dicom files.
        """
        for series_key, series_description_list in self.series_descriptions.items():
            if any(series in self.available_series_descriptions for series in series_description_list):
                pass
            else:
                self.update_series_descriptions(series_key)
                self.check_availability_of_given_series_description()

    def check_availability_of_given_series_uids(self) -> None:
        """
        Check availability of given series uids in the patient's dicom files.
        """
        for path_to_segmentation in self.paths_to_segmentations:
            if any(uid in path_to_segmentation for uid in self.available_series_uids):
                pass
            else:
                logging.warning(f"The given segmentation file name is {path_to_segmentation}. \nHowever, this file name"
                                f" does NOT contain any of the available series uids found in the patient's dicom "
                                f"files. \nAvailable series uids are {self.available_series_uids}.")

    def get_patient_dataset(self) -> PatientDataModel:
        """
        Get the patient dataset.

        Returns
        -------
        patient_dataset : PatientDataModel
            A named tuple grouping the patient's data extracted from its dicom files and the patient's medical image
            segmentation data extracted from the segmentation files.
        """
        patient_data_context = PatientDataQueryContext(
            images_data=self.get_images_data(self._verbose),
            organs=self._organs,
            paths_to_segmentations=self._paths_to_segmentations,
            series_descriptions=self._series_descriptions
        )
        patient_dataset = patient_data_context.create_patient_data()

        if self._verbose:
            logging.debug("\nThe chosen patient data query strategy is called "
                          "'{patient_data_context.patient_data_query_strategy.name}'.")
            logging.info(f"\nA total of {len(patient_dataset.data)} images were added to the patient dataset, namely:")

            for image_and_segmentation_data in patient_dataset.data:
                modality = image_and_segmentation_data.image.dicom_header.Modality
                series_description = image_and_segmentation_data.image.dicom_header.SeriesDescription
                segmentation = image_and_segmentation_data.segmentation
                image_segmentation_available = True if segmentation else False
                segmented_organs = list(segmentation.binary_label_maps.keys()) if segmentation else None

                logging.info(f"---> Series Description ({series_description})"
                             f"\n     ---> Modality : {modality}"
                             f"\n     ---> Image Segmentation available: {image_segmentation_available}")

                if image_segmentation_available:
                    logging.info(f"     ---> Segmented organs : {segmented_organs}")

        return patient_dataset



"""
    @file:              dicom_reader.py
    @Author:            Maxence Larose

    @Creation Date:     10/2021
    @Last modification: 01/2022

    @Description:       This file contains the DicomReader class which is used to read dicom files contained in a given
                        folder.
"""

import logging
from typing import Dict, List, NamedTuple, Set

import pydicom
import SimpleITK as sitk

from ...data_model import ImageDataModel
from ...utils import check_validity_of_given_path


class DicomReader:
    """
    A class used to read dicom files contained in a given folder.
    """

    class SeriesData(NamedTuple):
        """
        Series description namedtuple to simplify management of values.
        """
        series_description: str
        paths_to_dicoms_from_series: List[str]
        dicom_header: pydicom.dataset.FileDataset

    def __init__(
            self,
            path_to_dicom_folder: str
    ):
        """
        Constructor of the class DicomReader.

        Parameters
        ----------
        path_to_dicom_folder : str
            Path to the folder containing the patient dicom files.
        """
        super(DicomReader, self).__init__()

        check_validity_of_given_path(path=path_to_dicom_folder)
        self._path_to_dicom_folder = path_to_dicom_folder

    @staticmethod
    def get_dicom_header(
            path_to_dicom: str,
            show: bool = False
    ) -> pydicom.dataset.FileDataset:
        """
        Get a dicom header given the path to the dicom.

        Parameters
        ----------
        path_to_dicom : str
            The path to the dicom file.
        show : bool
            Show dicom header in console.

        Returns
        -------
        loaded_dicom : pydicom.dataset.FileDataset
            Loaded DICOM dataset.
        """
        loaded_dicom = pydicom.dcmread(path_to_dicom, stop_before_pixels=True)

        if show:
            logging.info(loaded_dicom)

        return loaded_dicom

    @property
    def __series_ids(self) -> List[str]:
        """
        Get all series IDs from a patient's dicom folder.

        Returns
        -------
        series_ids : List[str]
            All series IDs contained in a patient folder.
        """
        series_reader = sitk.ImageSeriesReader()
        series_ids = series_reader.GetGDCMSeriesIDs(self._path_to_dicom_folder)

        if not series_ids:
            raise FileNotFoundError(f"Given directory {self._path_to_dicom_folder} does not contain a DICOM series.")

        return series_ids

    @property
    def __series_data_dict(self) -> Dict[str, SeriesData]:
        """
        Get the series data from series IDs and the path to the patient's dicom folder.

        Returns
        -------
        series_data_dict : Dict[str, SeriesData]
            Dictionary of the data from the selected series.
        """
        series_data_dict: Dict[str, DicomReader.SeriesData] = {}
        all_patient_names: Set[str] = set()
        for idx, series_id in enumerate(self.__series_ids):
            series_reader = sitk.ImageSeriesReader()
            paths_to_dicoms_from_series = series_reader.GetGDCMSeriesFileNames(self._path_to_dicom_folder, series_id)

            path_to_first_dicom_of_series = paths_to_dicoms_from_series[0]
            loaded_dicom_header = self.get_dicom_header(path_to_dicom=path_to_first_dicom_of_series)
            all_patient_names.add(loaded_dicom_header.PatientName)

            series_data = self.SeriesData(
                series_description=loaded_dicom_header.SeriesDescription,
                paths_to_dicoms_from_series=paths_to_dicoms_from_series,
                dicom_header=loaded_dicom_header
            )
            series_data_dict[series_id] = series_data

        if len(all_patient_names) != 1:
            raise AssertionError(f"All dicom files in the same folder must belong to the same patient. This is not the "
                                 f"case for the patient whose data is currently being downloaded since the names "
                                 f"{all_patient_names} are found in his or her folder.")

        return series_data_dict

    @staticmethod
    def __get_3d_sitk_image_from_dicom_series(
            paths_to_dicoms_from_series: List[str],
    ) -> sitk.Image:
        """
        Get a 3D image array from a list of dicom paths associated to the same series.

        Parameters
        ----------
        paths_to_dicoms_from_series : List[str]
            A list of the paths to each dicom contained in the series.

        Returns
        -------
        image : sitk.Image
            3D SimpleITK image obtained from the series.
        """
        series_reader = sitk.ImageSeriesReader()
        series_reader.SetFileNames(paths_to_dicoms_from_series)

        image = series_reader.Execute()

        return image

    def get_dicom_headers(self, verbose: bool) -> List[pydicom.dataset.FileDataset]:
        """
        List of all the patient's dicom headers.

        Parameters
        ----------
        verbose : bool
            True to log/print some information else False.

        Returns
        -------
        dicom_headers : List[pydicom.dataset.FileDataset]
            A list of the patient's dicom headers.
        """
        dicom_headers = []
        for idx, series_data in enumerate(self.__series_data_dict.values()):
            if verbose:
                if idx == 0:
                    logging.info(f"Patient name : {series_data.dicom_header.PatientName}.")
                    logging.debug(f"\nPath to dicom folder : {self._path_to_dicom_folder}. \nThe following series "
                                  f"description are found in the patient's folder :")

                logging.debug(f"---> Series description: {series_data.series_description}")

            dicom_headers.append(series_data.dicom_header)

        return dicom_headers

    def get_images_data(self, verbose: bool) -> List[ImageDataModel]:
        """
        List of tuples containing 3D images array and their corresponding dicom header. Each element in the list
        corresponds to an image series.

        Parameters
        ----------
        verbose : bool
            True to log/print some information else False.

        Returns
        -------
        image_data_list : List[ImageDataModel]
            A list of the patient's images data.
        """
        images_data = []
        for idx, series_data in enumerate(self.__series_data_dict.values()):

            if verbose:
                if idx == 0:
                    logging.info(f"\nThe following series description are found in the patient's folder :")
                logging.info(f"---> Series description: {series_data.series_description}")

            image = self.__get_3d_sitk_image_from_dicom_series(
                paths_to_dicoms_from_series=series_data.paths_to_dicoms_from_series
            )

            image_data = ImageDataModel(
                dicom_header=series_data.dicom_header,
                simple_itk_image=image
            )

            images_data.append(image_data)

        return images_data
